/*******************************************************************************
#     sw-af: Software Autofocus for Logitech QuickCam Pro 9000                 #
# This package works with the Logitech UVC based webcams with the mjpeg feature#
# using the Logitech user control focus (mainly QC Pro 9000)                   #
#                                                                              #
#   Copyright (C) 2007 Alexander K. Seewald <alex@seewald.at>                  #
#   Includes GPL code (C) 2005 2006 2007 by Michel Xhaard                      #
#   Relicensed as LGPL in Feb. 2010 by permission of both authors              #
#                                                                              #
# This program is free software; you can redistribute it and/or modify         #
# it under the terms of the GNU Lesser General Public License as published by  #
# the Free Software Foundation; either version 2.1 of the License, or          #
# (at your option) any later version.                                          #
#                                                                              #
# This program is distributed in the hope that it will be useful,              #
# but WITHOUT ANY WARRANTY; without even the implied warranty of               #
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                #
# GNU Lesser General Public License for more details.                          #
#                                                                              #
# You should have received a copy of the GNU Lesser General Public License     #
# along with this program; if not, write to the Free Software                  #
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA    #
#                                                                              #
*******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <sys/select.h>
#include <unistd.h>
#include <jpeglib.h>
#include <linux/videodev.h>
#include <fcntl.h>
#include <errno.h>

#include <linux/uvcvideo.h>

#ifdef UVC_CONTROL_SET_CUR
#define UVC_DYN_CONTROLS
#endif

#include "SDL/SDL.h"

static const char version[] = "0.06";
static int debug=0;

// time in ms to change focus by 1 (estimated from range 0-255, assumed linear)
#define focusTime 14/10


#ifdef UVC_DYN_CONTROLS
#define V4L2_CID_FOCUS_LOGITECH 0x0A046D04
#else
#define V4L2_CID_FOCUS_LOGITECH V4L2_CID_FOCUS_ABSOLUTE
#endif

#define NB_BUFFER 2
#define DHT_SIZE 432

static unsigned char dht_data[DHT_SIZE] = {
  0xff, 0xc4, 0x00, 0x1f, 0x00, 0x00, 0x01, 0x05, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  0x00, 0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0xff, 0xc4, 0x00, 0xb5, 0x10, 0x00, 0x02,
  0x01, 0x03, 0x03, 0x02, 0x04, 0x03, 0x05, 0x05, 0x04, 0x04, 0x00, 0x00, 0x01, 0x7d, 0x01, 0x02, 0x03, 0x00, 0x04, 0x11,
  0x05, 0x12, 0x21, 0x31, 0x41, 0x06, 0x13, 0x51, 0x61, 0x07, 0x22, 0x71, 0x14, 0x32, 0x81, 0x91, 0xa1, 0x08, 0x23, 0x42,
  0xb1, 0xc1, 0x15, 0x52, 0xd1, 0xf0, 0x24, 0x33, 0x62, 0x72, 0x82, 0x09, 0x0a, 0x16, 0x17, 0x18, 0x19, 0x1a, 0x25, 0x26,
  0x27, 0x28, 0x29, 0x2a, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x53,
  0x54, 0x55, 0x56, 0x57, 0x58, 0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x73, 0x74, 0x75, 0x76, 0x77,
  0x78, 0x79, 0x7a, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a,
  0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3,
  0xc4, 0xc5, 0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe1, 0xe2, 0xe3, 0xe4,
  0xe5, 0xe6, 0xe7, 0xe8, 0xe9, 0xea, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa, 0xff, 0xc4, 0x00, 0x1f,
  0x01, 0x00, 0x03, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x01, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x01, 0x02,
  0x03, 0x04, 0x05, 0x06, 0x07, 0x08, 0x09, 0x0a, 0x0b, 0xff, 0xc4, 0x00, 0xb5, 0x11, 0x00, 0x02, 0x01, 0x02, 0x04, 0x04,
  0x03, 0x04, 0x07, 0x05, 0x04, 0x04, 0x00, 0x01, 0x02, 0x77, 0x00, 0x01, 0x02, 0x03, 0x11, 0x04, 0x05, 0x21, 0x31, 0x06,
  0x12, 0x41, 0x51, 0x07, 0x61, 0x71, 0x13, 0x22, 0x32, 0x81, 0x08, 0x14, 0x42, 0x91, 0xa1, 0xb1, 0xc1, 0x09, 0x23, 0x33,
  0x52, 0xf0, 0x15, 0x62, 0x72, 0xd1, 0x0a, 0x16, 0x24, 0x34, 0xe1, 0x25, 0xf1, 0x17, 0x18, 0x19, 0x1a, 0x26, 0x27, 0x28,
  0x29, 0x2a, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x43, 0x44, 0x45, 0x46, 0x47, 0x48, 0x49, 0x4a, 0x53, 0x54, 0x55, 0x56,
  0x57, 0x58, 0x59, 0x5a, 0x63, 0x64, 0x65, 0x66, 0x67, 0x68, 0x69, 0x6a, 0x73, 0x74, 0x75, 0x76, 0x77, 0x78, 0x79, 0x7a,
  0x82, 0x83, 0x84, 0x85, 0x86, 0x87, 0x88, 0x89, 0x8a, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97, 0x98, 0x99, 0x9a, 0xa2, 0xa3,
  0xa4, 0xa5, 0xa6, 0xa7, 0xa8, 0xa9, 0xaa, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7, 0xb8, 0xb9, 0xba, 0xc2, 0xc3, 0xc4, 0xc5,
  0xc6, 0xc7, 0xc8, 0xc9, 0xca, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7, 0xd8, 0xd9, 0xda, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,
  0xe8, 0xe9, 0xea, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7, 0xf8, 0xf9, 0xfa
};

struct vdIn {
  int fd;
  char *videodevice;
  char *status;
  char *pictName;
  struct v4l2_capability cap;
  struct v4l2_format fmt;
  struct v4l2_buffer buf;
  struct v4l2_requestbuffers rb;
  void *mem[NB_BUFFER];
  unsigned char *tmpbuffer;
  unsigned char *framebuffer;
  int isstreaming;
  int grabmethod;
  int width;
  int height;
  int fps;
  int formatIn;
  int formatOut;
  int framesizeIn;
  int signalquit;
  int toggleAvi;
  int getPict;
  int rawFrameCapture;
  /* raw frame capture */
  unsigned int fileCounter;
  /* raw frame stream capture */
  unsigned int rfsFramesWritten;
  unsigned int rfsBytesWritten;
  /* raw stream capture */
  FILE *captureFile;
  unsigned int framesWritten;
  unsigned int bytesWritten;
  int framecount;
  int recordstart;
  int recordtime;
};

#define SEARCH_RECURSIVE 0
#define SEARCH_LINEAR 1

double getFrameSharpnessValueWrapper(struct vdIn *videoIn);
double getFrameSharpnessValue(unsigned char *buf, int len);
int init_videoIn(struct vdIn *vd, char *device, int width, int height, int fps,
  int format, int grabmethod);
int v4l2SetControl(struct vdIn *vd, int control, int value);
int v4l2ResetControl(struct vdIn *vd, int control);
int close_v4l2(struct vdIn *vd);
int uvcGrab(struct vdIn *vd);
int is_huffman(unsigned char *buf);
int initDynCtrls(struct vdIn *vd);

static int init_v4l2(struct vdIn *vd);
static int isv4l2Control(struct vdIn *vd, int control,
       struct v4l2_queryctrl *queryctrl);
static int video_enable(struct vdIn *vd);
static int video_disable(struct vdIn *vd);

int get_picture_focus(unsigned char *buf,int size,char *format,int grp, int focus);

int main (int argc, char **argv) {
  char *videodevice = "/dev/video0";
  int grabmethod = 1;
  struct vdIn *videoIn;

  int width=640; int height=480; int fps=15;
  int format=V4L2_PIX_FMT_MJPEG;
  int stepSize=-1;
  int i;
  int dumpImages=FALSE;
  int pid = -1;
  int res;
  char *testImage = NULL;
  long curr;
  int doTiming=0;
  int tDelay=10;

  int searchMethod=-1;

  printf("%s software autofocus V%s (C) 2007 Alexander K. Seewald <alex@seewald.at>\n",argv[0],version);
  for (i=1; i < argc; i++) {
    if (argv[i] == NULL || *argv[i] != '-') {
      if (argv[i]==NULL) {
        printf("Ignoring bad argument NULL at position %d\n",i);
      } else {
        printf("Ignoring bad argument %s at position %d\n",argv[i],i);
      }
    } else if (strcmp(argv[i],"-d")==0) {
      if (i+1 >= argc) {
        fprintf(stderr,"No parameter specified with -d, aborting.\n");
        exit(1);
      } else {
        videodevice = strdup(argv[i+1]); i++;
      }
    } else if (strcmp(argv[i],"-r")==0) {
      if (searchMethod==SEARCH_LINEAR) { fprintf(stderr,"Both -r and -l cannot be given, aborting.\n"); exit(1); }
      searchMethod=SEARCH_RECURSIVE;
    } else if (strcmp(argv[i],"-l")==0) {
      if (searchMethod==SEARCH_RECURSIVE) { fprintf(stderr,"Both -r and -l cannot be given, aborting.\n"); exit(1); }
      searchMethod=SEARCH_LINEAR;
    } else if (strcmp(argv[i],"-t")==0) {
      if (i+1 >= argc) {
        fprintf(stderr,"No parameter specified with -t, aborting.\n");
        exit(1);
      } else {
        testImage = strdup(argv[i+1]);
        i++;
      }
    } else if (strcmp(argv[i],"-s")==0) {
      if (i+1 >= argc) {
        fprintf(stderr,"No parameter specified with -s, aborting.\n");
        exit(1);
      } else {
        stepSize = atoi(argv[i+1]);
        if (stepSize<=0 || stepSize>255) {
          fprintf(stderr,"Invalid step size, aborting.\n");
          exit(1);
        }
        i++;
      }
    } else if (strcmp(argv[i],"-D")==0) {
      dumpImages=TRUE; pid=getpid();
    } else if (strcmp(argv[i],"-h")==0) {
      fprintf(stderr,"Usage: %s [-h|-d [device]|-r|-f|-R|-l|-t [filename.jpg]|-s [step size]|-D]\n",argv[0]);
      fprintf(stderr,"-h  print this usage message.\n");
      fprintf(stderr,"-d  set capture device (default /dev/video0)\n");
      fprintf(stderr,"-r  use recursive search algorithm (default)\n");
      fprintf(stderr,"-f  30fps framerate instead of 15fps\n");
      fprintf(stderr,"    faster focus (<1.3s on a recent system)\n");
      fprintf(stderr,"-R  176x144 resolution instead of 640x480\n");
      fprintf(stderr,"    increases processing speed on slow machines\n");
      fprintf(stderr,"-t  output final sharp image for testing purposes\n");
      fprintf(stderr,"-l  use linear search algorithm\n");
      fprintf(stderr,"-T  output time until focus convergence\n");
      fprintf(stderr,"-s  set step size for linear search (default 8)\n");
      fprintf(stderr,"-D  dump images for debugging (enforces -l -s 8)\n");
      fprintf(stderr,"    Use this if you find a scene where the focus\n");
      fprintf(stderr,"    does not converge, and send me the test images.\n");
      exit(0);
    } else if (strcmp(argv[i],"-f")==0) {
      fps=30; tDelay=1;
    } else if (strcmp(argv[i],"-R")==0) {
      width=176; height=144;  
    } else if (strcmp(argv[i],"-T")==0) {
      doTiming=1;
    } else {
      fprintf(stderr,"Unknown parameter %s found, aborting.\n",argv[i]);
      exit(1);
    }
  }

  if (dumpImages) {
    if (searchMethod==SEARCH_RECURSIVE) {
      printf("Ignoring search method RECURSIVE because of -D\n");
    }
    if (stepSize!=-1 && stepSize!=8) {
      printf("Ignoring step size %d because of -D\n",stepSize);
    }
    searchMethod=SEARCH_LINEAR; stepSize=8;
    width=640; height=480; fps=15; tDelay=10;
  }

  if (searchMethod==-1) { searchMethod=SEARCH_RECURSIVE; }

  if (stepSize!=-1 && searchMethod==SEARCH_RECURSIVE) {
    printf("Step size %d was ignored since search method is RECURSIVE\n",stepSize);
  }

  printf("Focussing with search method %s, device %s", (searchMethod == SEARCH_RECURSIVE ? "RECURSIVE" : "LINEAR"), videodevice);

  if (searchMethod==SEARCH_LINEAR) {
    if (stepSize==-1) { stepSize=8; }
    printf(", step size %d",stepSize);
  }
  printf("\n");
  if (dumpImages) { printf("Images will be dumped to current directory as %d.[focus].jpg\n",pid); }

  // start initialization
  videoIn = (struct vdIn *) calloc (1, sizeof (struct vdIn));
  if (init_videoIn(videoIn, (char *) videodevice, width, height, fps, format, grabmethod) < 0) {
    exit (1);
  }

  // Initialize dynamic controls
  res=initDynCtrls(videoIn);
  if (res<0) { fprintf(stderr,"Initialization of dynamic control Focus failed.\n"); } else if (res==1) { printf("Dynamic control Focus already initialized, using existing control.\n"); } else if (res==2) { printf("UVC-V4L2 mapping for Focus already initialized, using existing mapping.\n"); }

  //Reset all camera controls
  v4l2ResetControl (videoIn, V4L2_CID_FOCUS_LOGITECH);
  v4l2ResetControl (videoIn, V4L2_CID_BRIGHTNESS);
  v4l2ResetControl (videoIn, V4L2_CID_SATURATION);
  v4l2ResetControl (videoIn, V4L2_CID_CONTRAST);
  v4l2ResetControl (videoIn, V4L2_CID_SHARPNESS);
  v4l2ResetControl (videoIn, V4L2_CID_GAIN);

  if (uvcGrab(videoIn)<0) { fprintf(stderr,"Getting first frame failed."); goto end; }

  if (doTiming) { curr=SDL_GetTicks(); }
  if (searchMethod==SEARCH_LINEAR) {
    int focus=0; double maxS=-1.0; int maxFocus=0; double sharpness; int t=0; int xi;
    v4l2SetControl(videoIn, V4L2_CID_FOCUS_LOGITECH, focus);
    t=(int)(255*focusTime)/(1000/fps)+1; if (t<2) { t=2; }
    for (xi=0; xi<t; xi++) {
      SDL_Delay(tDelay);
      if (uvcGrab(videoIn)<0) { fprintf(stderr,"Error grabbing focus sample %d\n",focus); goto end;}
    }
    while (focus<256) {
      t=(int)(stepSize*focusTime)/(1000/fps)+1; if (t<2) { t=2; }
      for (xi=0; xi<t; xi++) {
        SDL_Delay(tDelay);
        if (uvcGrab(videoIn)<0) { fprintf(stderr,"Error grabbing focus sample %d\n",focus); goto end;}
      }
      sharpness = getFrameSharpnessValueWrapper(videoIn);
      if (sharpness>maxS) { maxS=sharpness; maxFocus=focus; }
      if (dumpImages) get_picture_focus(videoIn->tmpbuffer,videoIn->buf.bytesused,NULL,pid,focus);
      focus+=stepSize;
      v4l2SetControl(videoIn, V4L2_CID_FOCUS_LOGITECH, focus);
    }
    v4l2SetControl(videoIn, V4L2_CID_FOCUS_LOGITECH, maxFocus);
    printf("Chosen focus=%d\n",maxFocus);
    if (doTiming) { curr=SDL_GetTicks()-curr; }
    if (testImage!=NULL) {
      t=(int)(abs(maxFocus-(focus-stepSize))*focusTime)/(1000/fps)+1; if (t<2) { t=2; }
      for (xi=0; xi<t; xi++) {
        SDL_Delay(tDelay);
        if (uvcGrab(videoIn)<0) { fprintf(stderr,"Error grabbing focus sample\n"); goto end;}
      }
    }
  } else {
    int left=0; int right=256; int middle=(left+right)/2; double leftVal=0.0; double rightVal=0.0; int lastF=255; int t=0; int xi;
    int focusNow; int leftFirst=1;
    while ((right-left)>2) {
      middle=(left+right)/2;
      if (leftFirst) { focusNow=(left+middle)/2; } else { focusNow=(right+middle)/2; }
      v4l2SetControl(videoIn, V4L2_CID_FOCUS_LOGITECH, focusNow);
      t=(int)((abs(lastF-focusNow))*focusTime)/(1000/fps)+1; if (t<2) { t=2; }
      lastF=focusNow;
      for (xi=0; xi<t; xi++) {
        SDL_Delay(tDelay);
        if (uvcGrab(videoIn)<0) { fprintf(stderr,"Error grabbing focus sample\n"); goto end;}
      }

      if (!leftFirst) { focusNow=(left+middle)/2; } else { focusNow=(right+middle)/2; }
       if (leftFirst) { leftVal=getFrameSharpnessValueWrapper(videoIn); } else { rightVal=getFrameSharpnessValueWrapper(videoIn); }
      v4l2SetControl(videoIn, V4L2_CID_FOCUS_LOGITECH, focusNow);
      t=(int)((abs(lastF-(right+middle)/2))*focusTime)/(1000/fps)+1; if (t<2) { t=2; }
      lastF=focusNow;
      for (xi=0; xi<t; xi++) {
        SDL_Delay(tDelay);
        if (uvcGrab(videoIn)<0) { fprintf(stderr,"Error grabbing focus sample\n"); goto end;}
      }
       if (!leftFirst) { leftVal=getFrameSharpnessValueWrapper(videoIn); } else { rightVal=getFrameSharpnessValueWrapper(videoIn); }
      if (leftVal>rightVal) { right=middle; } else { left=middle; }
      leftFirst=abs((left+middle)/2-lastF)<abs((right+middle)/2-lastF);
    }
    if (doTiming) { curr=SDL_GetTicks()-curr; }
    v4l2SetControl(videoIn, V4L2_CID_FOCUS_LOGITECH, (left+right)/2);
    printf("Chosen focus=%d\n",(left+right)/2);
    if (testImage!=NULL) {
      t=(int)(abs((left+right)/2-lastF)*focusTime)/(1000/fps)+1; if (t<2) { t=2; }
      for (xi=0; xi<t; xi++) {
        SDL_Delay(tDelay);
        if (uvcGrab(videoIn)<0) { fprintf(stderr,"Error grabbing focus sample\n"); goto end;}
      }
    }
  }
  if (testImage!=NULL) {
    SDL_Delay(tDelay);
    if (uvcGrab(videoIn)<0) { fprintf(stderr,"Error grabbing focus sample\n"); goto end;}
    get_picture_focus(videoIn->tmpbuffer,videoIn->buf.bytesused,testImage,0,0);
  }
  if (doTiming) { printf("Time to focus: %ld ms\n",curr); }

end:
  close_v4l2(videoIn);
  free(videoIn);

  return 0;
}

double getFrameSharpnessValueWrapper(struct vdIn *videoIn) {
  unsigned char *tmp_buf = malloc(videoIn->buf.bytesused+DHT_SIZE);
  unsigned char *ptdeb, *ptcur;
  ptdeb=ptcur=videoIn->tmpbuffer;
  while (((ptcur[0] << 8) | ptcur[1]) != 0xffc0) ptcur++;
  int sizein = ptcur-ptdeb;
  memcpy(tmp_buf,ptdeb,sizein);
  memcpy(tmp_buf+sizein,dht_data,DHT_SIZE);
  memcpy(tmp_buf+sizein+DHT_SIZE,ptcur,videoIn->buf.bytesused-sizein);
  double sv = getFrameSharpnessValue(tmp_buf,videoIn->buf.bytesused+DHT_SIZE);
  free(tmp_buf);
  if (sv==-1.0) {
    if (uvcGrab(videoIn)<0) { fprintf(stderr,"uvcGrab problem in getFrameSharpnessValueWrapper: %d",errno); exit(-1); }
    return getFrameSharpnessValueWrapper(videoIn);
  }
  return sv;
}


int init_videoIn(struct vdIn *vd, char *device, int width, int height, int fps, int format, int grabmethod)
{
  if (vd == NULL || device == NULL)
    return -1;
  if (width == 0 || height == 0)
    return -1;
  if (grabmethod < 0 || grabmethod > 1)
    grabmethod = 1;    //mmap by default;

  vd->videodevice = NULL;
  vd->status = NULL;
  vd->pictName = NULL;
  vd->videodevice = (char *) calloc(1, 16 * sizeof(char));
  vd->status = (char *) calloc(1, 100 * sizeof(char));
  vd->pictName = (char *) calloc(1, 80 * sizeof(char));
  snprintf(vd->videodevice, 12, "%s", device);
//  printf("video %s \n", vd->videodevice);
  vd->toggleAvi = 0;
  vd->recordtime = 0;
  vd->framecount = 0;
  vd->recordstart = 0;
  vd->getPict = 0;
  vd->signalquit = 1;
  vd->width = width;
  vd->height = height;
  vd->fps = fps;
  vd->formatIn = format;
  vd->grabmethod = grabmethod;
  vd->fileCounter = 0;
  vd->rawFrameCapture = 0;
  vd->rfsBytesWritten = 0;
  vd->rfsFramesWritten = 0;
  vd->captureFile = NULL;
  vd->bytesWritten = 0;
  vd->framesWritten = 0;
  if (init_v4l2(vd) < 0) {
    printf(" Init v4L2 failed !! exit fatal \n");
    goto error;;
  }
  /* alloc a temp buffer to reconstruct the pict */
  vd->framesizeIn = (vd->width * vd->height << 1);
  switch (vd->formatIn) {
    case V4L2_PIX_FMT_MJPEG:
      vd->tmpbuffer = (unsigned char *) calloc(1, (size_t) vd->framesizeIn);
      if (!vd->tmpbuffer)
        goto error;
      vd->framebuffer = (unsigned char *) calloc(1, (size_t) vd->width * (vd->height + 8) * 2);
      break;
    case V4L2_PIX_FMT_YUYV:
      vd->framebuffer = (unsigned char *) calloc(1, (size_t) vd->framesizeIn);
      break;
    default:
      fprintf(stderr," should never arrive exit fatal !!\n");
      goto error;
      break;
  }
  if (!vd->framebuffer)
    goto error;
  return 0;

error:
  free(vd->videodevice);
  free(vd->status);
  free(vd->pictName);
  close(vd->fd);
  return -1;
}

static int init_v4l2(struct vdIn *vd)
{
  int i;
  int ret = 0;

  if ((vd->fd = open(vd->videodevice, O_RDWR)) == -1) {
    perror("ERROR opening V4L interface \n");
    exit(1);
  }

  memset(&vd->cap, 0, sizeof(struct v4l2_capability));
  ret = ioctl(vd->fd, VIDIOC_QUERYCAP, &vd->cap);
  if (ret < 0) {
    printf("Error opening device %s: unable to query device.\n",
         vd->videodevice);
    goto fatal;
  }

  if ((vd->cap.capabilities & V4L2_CAP_VIDEO_CAPTURE) == 0) {
    printf("Error opening device %s: video capture not supported.\n",
         vd->videodevice);
    goto fatal;;
  }
  if (vd->grabmethod) {
    if (!(vd->cap.capabilities & V4L2_CAP_STREAMING)) {
      printf("%s does not support streaming i/o\n", vd->videodevice);
      goto fatal;
    }
  } else {
    if (!(vd->cap.capabilities & V4L2_CAP_READWRITE)) {
      printf("%s does not support read i/o\n", vd->videodevice);
      goto fatal;
    }
  }
  /* set format in */
  memset(&vd->fmt, 0, sizeof(struct v4l2_format));
  vd->fmt.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  vd->fmt.fmt.pix.width = vd->width;
  vd->fmt.fmt.pix.height = vd->height;
  vd->fmt.fmt.pix.pixelformat = vd->formatIn;
  vd->fmt.fmt.pix.field = V4L2_FIELD_ANY;
  ret = ioctl(vd->fd, VIDIOC_S_FMT, &vd->fmt);
  if (ret < 0) {
    printf("Unable to set format: %d.\n", errno);
    goto fatal;
  }

  if ((vd->fmt.fmt.pix.width != vd->width) ||
  (vd->fmt.fmt.pix.height != vd->height)) {
    printf(" format asked unavailable get width %d height %d \n",
         vd->fmt.fmt.pix.width, vd->fmt.fmt.pix.height);
    vd->width = vd->fmt.fmt.pix.width;
    vd->height = vd->fmt.fmt.pix.height;
    /* look the format is not part of the deal ??? */
    //vd->formatIn = vd->fmt.fmt.pix.pixelformat;
  }
    
  /* set framerate */
  struct v4l2_streamparm* setfps;  
  setfps=(struct v4l2_streamparm *) calloc(1, sizeof(struct v4l2_streamparm));
  memset(setfps, 0, sizeof(struct v4l2_streamparm));
  setfps->type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  setfps->parm.capture.timeperframe.numerator=1;
  setfps->parm.capture.timeperframe.denominator=vd->fps;
  ret = ioctl(vd->fd, VIDIOC_S_PARM, setfps); 
       
  /* request buffers */
  memset(&vd->rb, 0, sizeof(struct v4l2_requestbuffers));
  vd->rb.count = NB_BUFFER;
  vd->rb.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  vd->rb.memory = V4L2_MEMORY_MMAP;

  ret = ioctl(vd->fd, VIDIOC_REQBUFS, &vd->rb);
  if (ret < 0) {
    printf("Unable to allocate buffers: %d.\n", errno);
    goto fatal;
  }
  /* map the buffers */
  for (i = 0; i < NB_BUFFER; i++) {
    memset(&vd->buf, 0, sizeof(struct v4l2_buffer));
    vd->buf.index = i;
    vd->buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    vd->buf.memory = V4L2_MEMORY_MMAP;
    ret = ioctl(vd->fd, VIDIOC_QUERYBUF, &vd->buf);
    if (ret < 0) {
      printf("Unable to query buffer (%d).\n", errno);
      goto fatal;
    }
    if (debug)
      printf("length: %u offset: %u\n", vd->buf.length,
    vd->buf.m.offset);
    vd->mem[i] = mmap(0 /* start anywhere */ ,
    vd->buf.length, PROT_READ, MAP_SHARED, vd->fd,
    vd->buf.m.offset);
    if (vd->mem[i] == MAP_FAILED) {
      printf("Unable to map buffer (%d)\n", errno);
      goto fatal;
    }
    if (debug)
      printf("Buffer mapped at address %p.\n", vd->mem[i]);
  }
  /* Queue the buffers. */
  for (i = 0; i < NB_BUFFER; ++i) {
    memset(&vd->buf, 0, sizeof(struct v4l2_buffer));
    vd->buf.index = i;
    vd->buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    vd->buf.memory = V4L2_MEMORY_MMAP;
    ret = ioctl(vd->fd, VIDIOC_QBUF, &vd->buf);
    if (ret < 0) {
      printf("Unable to queue buffer (%d).\n", errno);
      goto fatal;;
    }
  }
  return 0;

fatal:
  return -1;
}

int v4l2SetControl(struct vdIn *vd, int control, int value)
{
  struct v4l2_control control_s;
  struct v4l2_queryctrl queryctrl;
  int min, max, step, val_def;
  int err;
  if (isv4l2Control(vd, control, &queryctrl) < 0)
    return -1;
  min = queryctrl.minimum;
  max = queryctrl.maximum;
  step = queryctrl.step;
  val_def = queryctrl.default_value;
  if ((value >= min) && (value <= max)) {
    control_s.id = control;
    control_s.value = value;
    if ((err = ioctl(vd->fd, VIDIOC_S_CTRL, &control_s)) < 0) {
      printf("ioctl set control error\n");
      return -1;
    }
  }
  return 0;
}

int initDynCtrls(struct vdIn *vd) {
#ifdef UVC_DYN_CONTROLS
#define UVC_GUID_LOGITECH_MOTOR_CONTROL {0x82, 0x06, 0x61, 0x63, 0x70, 0x50, 0xab, 0x49, 0xb8, 0xcc, 0xb3, 0x85, 0x5e, 0x8d, 0x22, 0x56}
#define XU_MOTORCONTROL_FOCUS 3

  int err;
  static struct uvc_xu_control_info ci = {
    .entity = UVC_GUID_LOGITECH_MOTOR_CONTROL,
    .selector = XU_MOTORCONTROL_FOCUS,
    .index = 2,
    .size = 6,
    .flags = UVC_CONTROL_SET_CUR|UVC_CONTROL_GET_MIN|UVC_CONTROL_GET_MAX|UVC_CONTROL_GET_DEF
  };

  errno=0;
  if ((err=ioctl(vd->fd, UVCIOC_CTRL_ADD, &ci)) < 0) {
    if (errno!=EEXIST) {
      printf("uvcioc ctrl add error: errno=%d (retval=%d)\n",errno,err);
      return -1;
    } else {
      return 1; // control exists
    }
  }

  static struct uvc_xu_control_mapping cm = {
    .id = V4L2_CID_FOCUS_LOGITECH,
    .name = "Focus",
    .entity = UVC_GUID_LOGITECH_MOTOR_CONTROL,
    .selector = XU_MOTORCONTROL_FOCUS,
    .size=8,
    .offset=0,
    .v4l2_type=V4L2_CTRL_TYPE_INTEGER,
    .data_type=UVC_CTRL_DATA_TYPE_UNSIGNED
  };

  errno=0;
  if ((err=ioctl(vd->fd, UVCIOC_CTRL_MAP, &cm)) < 0) {
    if (errno!=EEXIST) {
      printf("uvcioc ctrl map error: errno=%d (retval=%d)\n",errno,err);
      return -1;
    } else {
      return 2; // mapping exists
    }
  }
  return 0;

#else
  return 0;
#endif
}


int v4l2ResetControl(struct vdIn *vd, int control) {
  struct v4l2_control control_s;
  struct v4l2_queryctrl queryctrl;
  int val_def;
  int err;
  if (isv4l2Control(vd, control, &queryctrl)<0)
    return -1;
  val_def = queryctrl.default_value;
  control_s.id = control;
  control_s.value = val_def;
  if ((err = ioctl(vd->fd, VIDIOC_S_CTRL, &control_s)) < 0) {
    printf("ioctl reset control error\n");
    return -1;
  }

  return 0;
}

/* return >= 0 ok otherwhise -1 */
static int isv4l2Control(struct vdIn *vd, int control, struct v4l2_queryctrl *queryctrl) {
  int err =0;
  queryctrl->id = control;
  if ((err= ioctl(vd->fd, VIDIOC_QUERYCTRL, queryctrl)) < 0) {
    printf("ioctl querycontrol error %d \n",errno);
  } else if (queryctrl->flags & V4L2_CTRL_FLAG_DISABLED) {
    printf("control %s disabled \n", (char *) queryctrl->name);
  } else if (queryctrl->flags & V4L2_CTRL_TYPE_BOOLEAN) {
    return 1;
  } else if (queryctrl->type & V4L2_CTRL_TYPE_INTEGER) {
    return 0;
  } else {
    printf("control %s unsupported  \n", (char *) queryctrl->name);
  }
  return -1;
}


int close_v4l2(struct vdIn *vd) {
  if (vd->isstreaming)
    video_disable(vd);
  if (vd->tmpbuffer)
    free(vd->tmpbuffer);
  vd->tmpbuffer = NULL;
  free(vd->framebuffer);
  vd->framebuffer = NULL;
  free(vd->videodevice);
  free(vd->status);
  free(vd->pictName);
  vd->videodevice = NULL;
  vd->status = NULL;
  vd->pictName = NULL;
  return 0;
}

static int video_enable(struct vdIn *vd) {
  int type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  int ret;

  ret = ioctl(vd->fd, VIDIOC_STREAMON, &type);
  if (ret < 0) {
    printf("Unable to %s capture: %d.\n", "start", errno);
    return ret;
  }
  vd->isstreaming = 1;
  return 0;
}


static int video_disable(struct vdIn *vd)
{
  int type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  int ret;

  ret = ioctl(vd->fd, VIDIOC_STREAMOFF, &type);
  if (ret < 0) {
    printf("Unable to %s capture: %d.\n", "stop", errno);
    return ret;
  }
  vd->isstreaming = 0;
  return 0;
}


int uvcGrab(struct vdIn *vd) {
#define HEADERFRAME1 0xaf
  int ret;

  if (!vd->isstreaming)
    if (video_enable(vd))
      goto err;
  memset(&vd->buf, 0, sizeof(struct v4l2_buffer));
  vd->buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
  vd->buf.memory = V4L2_MEMORY_MMAP;
  ret = ioctl(vd->fd, VIDIOC_DQBUF, &vd->buf);
  if (ret < 0) {
    printf("Unable to dequeue buffer (%d).\n", errno);
    goto err;
  }

  switch (vd->formatIn) {
    case V4L2_PIX_FMT_MJPEG:
      if(vd->buf.bytesused <= HEADERFRAME1) {  /* Prevent crash on empty image */
        printf("Ignoring empty buffer ...\n");
        return 0;
      }
      memcpy(vd->tmpbuffer, vd->mem[vd->buf.index],vd->buf.bytesused);
      if (debug)
        printf("bytes in used %d \n", vd->buf.bytesused);
      break;
    default:
      goto err;
      break;
  }
  ret = ioctl(vd->fd, VIDIOC_QBUF, &vd->buf);
  if (ret < 0) {
    printf("Unable to requeue buffer (%d).\n", errno);
    goto err;
  }

  return 0;

err:
  vd->signalquit = 0;
  return -1;
}

int get_picture_focus(unsigned char *buf,int size,char *format,int grp, int focus) {
  FILE *file;
  unsigned char *ptdeb,*ptcur = buf;
  int sizein;
  char *name = NULL;
  name = calloc(80,1);
  if (format==NULL) { sprintf(name,"%d.%03d.jpg",grp,focus); }
  else { sprintf(name,format,grp,focus); }
  file = fopen(name, "wb");
  if (file != NULL) {
    if(!is_huffman(buf)){
      ptdeb = ptcur = buf;
      while (((ptcur[0] << 8) | ptcur[1]) != 0xffc0)
        ptcur++;
      sizein = ptcur-ptdeb;
      fwrite(buf, sizein, 1, file);
      fwrite(dht_data, DHT_SIZE, 1, file);
      fwrite(ptcur,size-sizein,1,file); 
    } else {
      fwrite(ptcur,size,1,file); /* ptcur was uninit -wsr */
    }     
    fclose(file);
  }
  if(name)
    free(name);
  return 0;    
}

int is_huffman(unsigned char *buf)
{
  unsigned char *ptbuf;
  int i = 0;
  ptbuf = buf;
  while (((ptbuf[0] << 8) | ptbuf[1]) != 0xffda){  
    if(i++ > 2048) 
      return 0;
    if(((ptbuf[0] << 8) | ptbuf[1]) == 0xffc4)
      return 1;
    ptbuf++;
  }
  return 0;
}
